/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Implements the vrpsx_allocator_t interface by combining a fixed-sized stack
 * with a delegate allocator for requests not satisfied by that stack.
 */

#define _POSIX_C_SOURCE 200809L

#include "hybrid_allocator.h"
#include <assert.h>


/* Expands to an lvalue for the trailer just before PTR.
 *
 * Trailers are used only in "multiplex" mode.
 * The trailer at the end of an allocation combines:
 *
 *     - a pointer to the start of that allocation, with
 *
 *     - a flag to indicate whether the *following* allocation remains in use,
 *       except that the flag in the final trailer wraps around, indicating
 *       whether the *first* allocation remains in use.
 *
 * Be sure to keep in mind that flag in a trailer applies to the following
 * allocation, but that the pointer applies to the previous allocation.
 *
 * We always merge freed final allocations with trailing free space,
 * so between method calls the in-use flag that applies to the final
 * allocation should be set.
 *
 * The in-use flag is combined with the pointer by overwriting the
 * low bit of the pointer.  Thanks to alignment restrictions, that
 * low bit would otherwise be zero, so we lose no information.
 *
 * This macro evaluates its argument exactly once.
 */
#define TRAILER(PTR)  (((char **)(PTR))[-1])

/* Given the value of a trailer, extracts the flag that indicates whether the
 * following allocation is still in use.  AS A SPECIAL CASE, this flag in the
 * final trailer applies to the first allocation.  Always 0 or 1.
 *
 * Evaluates its argument exactly once.
 */
#define IN_USE(TRAILER)  ((vrpsx_uintptr_t)(TRAILER) & 1)

/* Given the value of a trailer, returns a trailer with the
 * same previous allocation pointer but the in-use flag set to 1.
 *
 * Evaluates its argument exactly once.
 */
#define FORCE_USE(TRAILER)  ((char *)((vrpsx_uintptr_t)(TRAILER) | 1))

/* Given the value of a trailer, extracts a pointer to the start of
 * the allocation that it trails.  (There are more efficient ways to
 * do the same thing if you already know the value of the low bit.)
 *
 * Evaluates its argument exactly once.
 */
#define PREVIOUS(TRAILER)                                       \
    ((char *)((vrpsx_uintptr_t)(TRAILER) & (~ (vrpsx_uintptr_t)1)))


hybrid_allocator_vtbl_t const empty_hybrid_allocator_vtbl;
static hybrid_allocator_vtbl_t const singleton_hybrid_allocator_vtbl;
static hybrid_allocator_vtbl_t const multiplex_hybrid_allocator_vtbl;


static void empty_hybrid_allocator_free_memory(
    hybrid_allocator_t *obj,
    void *memory)
{
    /* The stack is empty, so this block must be from the delegate allocator.
     */
    VRPSX_CALL1(obj->delegate, free_memory, memory);
}

static void *empty_hybrid_allocator_allocate_memory(
    hybrid_allocator_t *obj,
    size_t size)
{
    char *begin;

    begin = obj->begin;
    if (size <= obj->end - begin) {
        /* The stack is big enough.  Transition to "singleton" state.
         */
        obj->unused = begin + size;
        obj->vtbl = &singleton_hybrid_allocator_vtbl;
        return begin;
    } else {
        /* Too big for stack, so delegate.  Stack remains empty.
         */
        return VRPSX_CALL1(obj->delegate, allocate_memory, size);
    }
}

static void *empty_hybrid_allocator_allocate_available(
    hybrid_allocator_t *obj,
    size_t *out_size)
{
    /* Allocate all stack space to one allocation.
     * Transition to "singleton" state.
     */
    char *begin = obj->begin;
    char *end = obj->end;
    obj->unused = end;
    obj->vtbl = &singleton_hybrid_allocator_vtbl;
    *out_size = end - begin;
    return begin;
}

static size_t empty_hybrid_allocator_shrink_final(
    hybrid_allocator_t *obj,
    size_t new_maximum_size)
{
    /* No final allocation to shrink.
     */
    return 0;
}


static void singleton_hybrid_allocator_free_memory(
    hybrid_allocator_t *obj,
    void *memory)
{
    if (memory == obj->begin) {
        /* Allocation was on stack.  Transition to "empty" state.
         */
        obj->vtbl = &empty_hybrid_allocator_vtbl;
    } else {
        /* Allocation came from delegate allocator.  Stay in "singleton" state.
         */
        VRPSX_CALL1(obj->delegate, free_memory, memory);
    }
}

static void *singleton_hybrid_allocator_allocate_memory(
    hybrid_allocator_t *obj,
    size_t size)
{
    char *begin, *oldUnused, *newUnused;
    size_t upgradeCost, delta;

    /* Compute the amount by which we would have to expand the first allocation
     * in order to support multiple allocations: the first trailer and padding.
     */
    begin = obj->begin;
    oldUnused = obj->unused;
    upgradeCost = sizeof(char *)                                 /* trailer */
            + ((- sizeof(char *) - (size_t)(oldUnused - begin))  /* +padding */
                    & (VRPSX_MAXIMUM_ALIGNMENT - 1));

    /* Compute the difference between the old and new values of (obj->unused):
     * (upgradeCost) plus the client-visible part of the second allocation
     * plus second trailer and any padding required for alignment purposes.
     *
     * If arithmetic overflow occurs, then it will wrap around according to the
     * rules of modular arithmetic because size_t is unsigned.  In particular,
     * there will be no hardware exception, and we will have (delta < size).
     */
    delta = size + sizeof(char *);
    delta += (- delta) & (VRPSX_MAXIMUM_ALIGNMENT - 1);
    delta += upgradeCost;

    /* We want the effect of allocating from the stack if there was no overflow
     * in the computation of (delta) and the stack has enough available space.
     *
     * But neither of those two checks has side effects, and therefore
     * we may reverse their chronological order in order to very slightly
     * speed the case in which we end up delegating.
     */
    if (delta <= (obj->end - oldUnused) && size < delta) {
        newUnused = oldUnused + delta;      /* find end of 2nd block */
        oldUnused += upgradeCost;           /* find new end of 1st block */
        TRAILER(oldUnused) = begin + 1;     /* +1 because 2nd block is in use*/
        TRAILER(newUnused) = oldUnused + 1; /* +1 because 1st block is in use*/
        obj->unused = newUnused;
        obj->vtbl = &multiplex_hybrid_allocator_vtbl; /* enter "multiplex" */
        return oldUnused;
    } else {
        /* Insufficient stack, so delegate and stay in "singleton" state.
         */
        return VRPSX_CALL1(obj->delegate, allocate_memory, size);
    }
}

static void *singleton_hybrid_allocator_allocate_available(
    hybrid_allocator_t *obj,
    size_t *out_size)
{
    char *begin, *oldUnused, *newUnused;
    size_t upgradeCost;

    /* Compute the amount by which we would have to expand the first allocation
     * in order to support multiple allocations: the first trailer and padding.
     */
    begin = obj->begin;
    oldUnused = obj->unused;
    upgradeCost = sizeof(char *)                                 /* trailer */
            + ((- sizeof(char *) - (size_t)(oldUnused - begin))  /* +padding */
                    & (VRPSX_MAXIMUM_ALIGNMENT - 1));

    /* Once the upgrade cost is paid, we need at least
     * VRPSX_MAXIMUM_ALIGNMENT more bytes for the new allocation,
     * even if its client-visible size were zero bytes.
     */
    newUnused = obj->end;
    if (upgradeCost + VRPSX_MAXIMUM_ALIGNMENT <= newUnused - oldUnused) {
        oldUnused += upgradeCost;           /* find new end of 1st block */
        TRAILER(oldUnused) = begin + 1;     /* +1 because 2nd block is in use*/
        TRAILER(newUnused) = oldUnused + 1; /* +1 because 1st block is in use*/
        obj->unused = newUnused;
        obj->vtbl = &multiplex_hybrid_allocator_vtbl; /* enter "multiplex" */
        *out_size = newUnused - sizeof(char *) - oldUnused;
        return oldUnused;
    } else {
        /* This operation is supposed to allocate stack space, never delegate.
         */
        *out_size = 0;
        return NULL;
    }
}

static size_t singleton_hybrid_allocator_shrink_final(
    hybrid_allocator_t *obj,
    size_t new_maximum_size)
{
    char *begin = obj->begin;
    size_t old_size = obj->unused - begin;

    if (old_size <= new_maximum_size) {
        return old_size;
    }

    obj->unused = begin + new_maximum_size;
    return new_maximum_size;
}


static void multiplex_hybrid_allocator_free_memory(
    hybrid_allocator_t *obj,
    void *memory)
{
    char *begin, *unused, *oldLastTrailer, *previousUnused, *trailer;

    begin = obj->begin;
    if (begin <= (char *)memory && (char *)memory < (unused = obj->unused)) {
        /* Allocation to be freed lies within stack.
         */
        oldLastTrailer = TRAILER(unused);
        previousUnused = PREVIOUS(oldLastTrailer);
        if (memory == previousUnused) {
            /* We are freeing the most recent allocation, so merge it
             * with trailing free space.  But also merge any adjacent
             * previously-freed allocations.
             *
             * Every time we start to execute the body of the loop:
             *
             *     memory: start of final allocation--after prior merging
             *
             *     previousUnused: start of free stack--after prior merging
             *
             * While initializing "memory" we subtract 1 because we know
             * that the allocation being freed had its in-use flag set.
             */
            trailer = TRAILER(memory);
            memory = trailer - 1;
            for (;;) {
                if (memory == begin) {
                    if (IN_USE(oldLastTrailer)) {
                        /* Freed all except the first allocation, which
                         * remains in use.  Transition to "singleton" state.
                         */
                        obj->unused = previousUnused;
                        obj->vtbl = &singleton_hybrid_allocator_vtbl;
                    } else {
                        /* Freed all allocations.  Transition to "empty" state.
                         */
                        obj->vtbl = &empty_hybrid_allocator_vtbl;
                    }
                    return;
                } else {
                    trailer = TRAILER(memory);
                    if (! IN_USE(trailer)) {
                        previousUnused = memory;
                        memory = trailer; /* Just copy because flag is clear.*/
                    } else {
                        /* Found an allocation that remains in use.
                         *
                         * That allocation will become the new final
                         * allocation.  And because it is not the first
                         * allocation, we will stay in "multiplex" mode.
                         *
                         * The new final allocation's trailer should continue
                         * to hold a pointer to the start of that allocation,
                         * but its in-use flag no longer refers to some later
                         * allocation, but to the very first allocation.
                         */
                        TRAILER(previousUnused)
                                = memory + IN_USE(oldLastTrailer);
                        obj->unused = previousUnused;
                        return;
                    }
                }
            }
        } else {
            /* Later allocations remain in use and prevent us from
             * merging the allocation being freed with trailing free space.
             * Therefore we just clear the applicable in-use flag.
             */
            if (memory == begin) {
                /* In this case the freed allocation is the very first, and
                 * the trailer that holds its in-use flag is the final one.
                 */
                TRAILER(unused) = previousUnused;
            } else {
                /* This allocation was not the first, and therefore its
                 * in-use flag is in the trailer of the previous allocation.
                 */
                trailer = TRAILER(memory);
                TRAILER(memory) = PREVIOUS(trailer);
            }
        }
    } else {
        /* Allocation to be freed must have come from delegate allocator.
         */
        VRPSX_CALL1(obj->delegate, free_memory, memory);
    }
}

static void *multiplex_hybrid_allocator_allocate_memory(
    hybrid_allocator_t *obj,
    size_t size)
{
    char *oldUnused, *newUnused, *oldTrailer;
    size_t delta;

    /* Compute the difference between the old and new values of (obj->unused):
     * the client-visible part of the new allocation, plus the new trailer,
     * plus any padding required for alignment purposes.
     *
     * If arithmetic overflow occurs, then it will wrap around according to the
     * rules of modular arithmetic because size_t is unsigned.  In particular,
     * there will be no hardware exception, and we will have (delta < size).
     */
    delta = size + sizeof(char *);
    delta += (- delta) & (VRPSX_MAXIMUM_ALIGNMENT - 1);

    /* We want the effect of allocating from the stack if there was no overflow
     * in the computation of (delta) and the stack has enough available space.
     *
     * But neither of those two checks has side effects, and therefore
     * we may reverse their chronological order in order to very slightly
     * speed the case in which we end up delegating.
     */
    oldUnused = obj->unused;
    if (delta <= obj->end - oldUnused && size < delta) {
        newUnused = oldUnused + delta;   /* find end of new block */
        oldTrailer = TRAILER(oldUnused); /* save in-use flag for 1st block */
        TRAILER(oldUnused) = FORCE_USE(oldTrailer); /* new block is in-use */
        TRAILER(newUnused) = oldUnused + IN_USE(oldTrailer);
                                        /* initialize new trailer: start of new
                                         * block + in-use flag for 1st block */
        obj->unused = newUnused;        /* update free space */
        return oldUnused;
    } else {
        /* Insufficient stack, so delegate and stay in "multiplex" state.
         */
        return VRPSX_CALL1(obj->delegate, allocate_memory, size);
    }
}

static void *multiplex_hybrid_allocator_allocate_available(
    hybrid_allocator_t *obj,
    size_t *out_size)
{
    char *oldUnused, *newUnused, *oldTrailer;

    /* We need at least VRPSX_MAXIMUM_ALIGNMENT bytes for the new allocation.
     */
    oldUnused = obj->unused;
    newUnused = obj->end;
    if (VRPSX_MAXIMUM_ALIGNMENT <= newUnused - oldUnused) {
        oldTrailer = TRAILER(oldUnused);  /* save in-use flag for 1st block */
        TRAILER(oldUnused) = FORCE_USE(oldTrailer); /* new block is in-use */
        TRAILER(newUnused) = oldUnused + IN_USE(oldTrailer);
                                        /* initialize new trailer: start of new
                                         * block + in-use flag for 1st block */
        obj->unused = newUnused;        /* update free space */
        *out_size = newUnused - sizeof(char *) - oldUnused;
        return oldUnused;
    } else {
        /* This operation is supposed to allocate stack space, never delegate.
         */
        *out_size = 0;
        return NULL;
    }
}

static size_t multiplex_hybrid_allocator_shrink_final(
    hybrid_allocator_t *obj,
    size_t new_maximum_size)
{
    char *oldUnused = obj->unused;
    char *trailer = TRAILER(oldUnused);
    char *final = PREVIOUS(trailer);
    size_t old_size = oldUnused - sizeof(char *) - final;
    size_t adjusted_size = new_maximum_size
            + ((- sizeof(char *) - new_maximum_size)
                    & (VRPSX_MAXIMUM_ALIGNMENT - 1));
    char *newUnused;

    /* If after alignment considerations we would not change the size,
     * or overflow occurred because the new maximum is huge, quit now.
     */
    if (old_size <= adjusted_size || adjusted_size < new_maximum_size) {
        return old_size;
    }

    newUnused = final + adjusted_size + sizeof(char *);
    TRAILER(newUnused) = trailer;
    obj->unused = newUnused;
    return adjusted_size;
}


static vrpsx_allocator_end_t const allocator_end;

static hybrid_allocator_end_t const hybrid_allocator_end;

/* Methods when the fixed-size stack holds no allocations at all.
 */
hybrid_allocator_vtbl_t const empty_hybrid_allocator_vtbl = {
    &empty_hybrid_allocator_free_memory,
    &empty_hybrid_allocator_allocate_memory,
    &allocator_end,
    &empty_hybrid_allocator_allocate_available,
    &empty_hybrid_allocator_shrink_final,
    &hybrid_allocator_end
};

/* Methods when the fixed-size stack holds exactly one allocation.
 */
static hybrid_allocator_vtbl_t const singleton_hybrid_allocator_vtbl = {
    &singleton_hybrid_allocator_free_memory,
    &singleton_hybrid_allocator_allocate_memory,
    &allocator_end,
    &singleton_hybrid_allocator_allocate_available,
    &singleton_hybrid_allocator_shrink_final,
    &hybrid_allocator_end
};

/* Methods when the fixed-size stack holds two or more allocations.
 */
static hybrid_allocator_vtbl_t const multiplex_hybrid_allocator_vtbl = {
    &multiplex_hybrid_allocator_free_memory,
    &multiplex_hybrid_allocator_allocate_memory,
    &allocator_end,
    &multiplex_hybrid_allocator_allocate_available,
    &multiplex_hybrid_allocator_shrink_final,
    &hybrid_allocator_end
};
