/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Declares an implementation of the vrpsx_allocator_t interface
 * that combines a fixed-sized stack with a delegate allocator
 * for requests not satisfied by that stack.
 */

#ifndef HYBRID_ALLOCATOR_H
#define HYBRID_ALLOCATOR_H 1

#include "../interface/vrpsx_allocator.h"
#include "vrpsx_utility.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* Implements the vrpsx_allocator_t by combining a fixed-size stack
 * with a delegate allocator for requests not satisfied by that stack.
 *
 * That fixed-size stack space is a constructor argument,
 * not part of the implementation struct as such.
 *
 * If the delegate supports more than "singular allocation" (as
 * defined by vrpsx_allocator_t), then so does this allocator.
 *
 * The alignment guaranteed made by this allocator is the
 * greatest common divisor of VRPSX_MAXIMUM_ALIGNMENT and
 * the alignment guaranteed by the delegate allocator.
 *
 * You do NOT need to explicitly destroy this allocator if you take
 * appropriate steps to ensure that delegate allocations are freed.
 * Memory blocks allocated in the fixed-size stack are implicitly
 * freed when that stack is destroyed.  You may use the delegate
 * allocator directly to free memory blocks that are located outside
 * of the fixed-size stack.  You may find these convenient:
 *
 *     - HYBRID_ALLOCATOR_STACK_CONTAINS
 *
 *     - STACK_FOR_HYBRID_ALLOCATOR_CONTAINS
 *
 * Performance Note:
 *
 * Within the fixed-size stack freed blocks are not fully reclaimed until
 * all subsequently allocated blocks are also freed.  Therefore you may
 * benefit from freeing blocks in the reverse of their order of allocation.
 *
 * But if you do not, then the only cost is otherwise-unnecessary use of
 * the delegate allocator.  (We do not try to prevent that outcome because
 * doing so would turn the fixed-size stack into a general-purpose allocator,
 * at which point it would be no faster than the delegate allocator.)
 *
 * Implementation Details:
 *
 * Note that the implementation changes the "vtbl" member when convenient
 * in order to handle changes in behavior as the instance changes state.
 * Such changes are allowed--see "Method modification" in "vrpsx_system.h".
 *
 * The "vtbl" member chooses between three basic state:
 *
 *     empty: zero allocations in stack
 *
 *     singleton: one allocation in stack (without any "trailer", as would
 *                be needed in the "multiplex" state) followed by free space
 *
 *     multiplex: two or more allocations in stack, each followed
 *                by alignment padding and then a "trailer", with
 *                the final allocation followed by usable free space
 *
 * See the source file containing the implementation for further details.
 *
 * Implements: vrpsx_allocator_t.
 * Implementation Lifetime: dictated by context.
 * Thread Safety: requires synchronization, but is no safer than the
 *                delegate allocator even when access is synchronized.
 */
typedef struct hybrid_allocator_struct {
#define HYBRID_ALLOCATOR_MEMBERS(VTBL)                                  \
    VRPSX_ALLOCATOR_MEMBERS(VTBL)                                       \
                                                                        \
    /* This allocator is used when fixed-size stack is insufficient.    \
     */                                                                 \
    vrpsx_allocator_t *delegate;                                        \
                                                                        \
    /* The beginning of the fixed-size stack.                           \
     */                                                                 \
    char *begin;                                                        \
                                                                        \
    /* The end of the fixed-size stack, including free space.           \
     */                                                                 \
    char *end;                                                          \
                                                                        \
    /* Points to the first position just after the used part of the stack. \
     * NOTE: Has a garbage value if there are no stack allocations.     \
     */                                                                 \
    char *unused;

    HYBRID_ALLOCATOR_MEMBERS(struct hybrid_allocator_vtbl_struct)
} hybrid_allocator_t;

typedef struct hybrid_allocator_end_struct {
} hybrid_allocator_end_t;

typedef struct hybrid_allocator_vtbl_struct {
#define HYBRID_ALLOCATOR_VTBL_MEMBERS(TYPE)                             \
    VRPSX_ALLOCATOR_VTBL_MEMBERS(TYPE)                                  \
                                                                        \
    /* Allocates all remaining space in the fixed-size stack            \
     * and stores the size of that allocation in (*out_size)            \
     * (ignoring any previous value of (*out_size)).  But if there      \
     * is insufficient stack space for any allocation, even an          \
     * empty one, then stores 0 in (*out_size) and returns NULL.        \
     */                                                                 \
    void *(*allocate_available)(                                        \
        TYPE *allocator,                                                \
        size_t *out_size);                                              \
                                                                        \
    /* Shrinks the final stack allocation to the minimum of its         \
     * old size and (new_maximum_size) bytes without moving it,         \
     * possibly rounding up for alignment reasons, and returns          \
     * the new size, including any such rounding.  Returns 0            \
     * if there are no allocations in the stack.                        \
     */                                                                 \
    size_t (*shrink_final)(                                             \
        TYPE *allocator,                                                \
        size_t new_maximum_size);                                       \
                                                                        \
    /* End marker for documentation and to help detect initialization bugs. \
     */                                                                 \
    hybrid_allocator_end_t const *hybrid_allocator_end;

    HYBRID_ALLOCATOR_VTBL_MEMBERS(hybrid_allocator_t);
} hybrid_allocator_vtbl_t;


/* Constructs a new instance of hybrid_allocator_t that
 * makes use of [BEGIN, END) as its fixed-size stack and
 * delegates to DELEGATE when that stack does not suffice.
 *
 * Of course, [BEGIN, END) should never be used by DELEGATE.
 *
 * You do NOT need to explicitly destroy the constructed allocator;
 * see the comments for hybrid_allocator_struct.
 *
 * See Also: hybrid_allocator_construct_from_array, HYBRID_ALLOCATOR_STACK
 *
 * Evaluates every argument exactly once.
 */
#define hybrid_allocator_construct(OBJ, DELEGATE, BEGIN, END)  {        \
    extern hybrid_allocator_vtbl_t const empty_hybrid_allocator_vtbl;   \
    hybrid_allocator_t *obj_565708dc = (OBJ);                           \
    vrpsx_allocator_t *delegate_565708dc = (DELEGATE);                  \
    vrpsx_maximally_aligned_t *begin_565708dc = (BEGIN);                \
    vrpsx_maximally_aligned_t *end_565708dc = (END);                    \
                                                                        \
    obj_565708dc->vtbl = &empty_hybrid_allocator_vtbl;                  \
    obj_565708dc->delegate = delegate_565708dc;                         \
    obj_565708dc->begin = (char *)begin_565708dc;                       \
    obj_565708dc->end = (char *)end_565708dc;                           \
}

/* Similar to hybrid_allocator_construct
 * but you must pass an array identifier.
 *
 * See Also: hybrid_allocator_construct, HYBRID_ALLOCATOR_STACK
 *
 * Evaluates every argument exactly once, except for ARRAY.
 */
#define hybrid_allocator_construct_from_array(OBJ, DELEGATE, ARRAY)     \
    hybrid_allocator_construct(OBJ, DELEGATE, ARRAY, VRPSX_ARRAY_END(ARRAY))

/* Declares an array with element type vrpsx_maximally_aligned_t
 * having the minimum dimension required to reach at least BYTES
 * bytes.  You specify the name of the array.  Please append ";".
 *
 * See Also: hybrid_allocator_construct,
 *           hybrid_allocator_construct_from_array
 */
#define HYBRID_ALLOCATOR_STACK(NAME, BYTES)             \
    vrpsx_maximally_aligned_t NAME [                    \
        (BYTES + sizeof(vrpsx_maximally_aligned_t) - 1) \
        / sizeof(vrpsx_maximally_aligned_t) ]

/* Given an address (MEMORY) inside a memory block currently allocated by
 * hybrid allocator (HYBRID_ALLOCATOR), expands to a nonzero expression
 * if and only if (MEMORY) was allocated in the stack of (HYBRID_ALLOCATOR),
 * as opposed to having been allocated by the delegate allocator.
 *
 * STACK_FOR_HYBRID_ALLOCATOR_CONTAINS may be faster when the
 * stack used by the hybrid allocator is currently in scope.
 *
 * WARNING: This macro evaluates its arguments more than once.
 */
#define HYBRID_ALLOCATOR_STACK_CONTAINS(HYBRID_ALLOCATOR, MEMORY)       \
    ((HYBRID_ALLOCATOR)->begin <= (char *)(MEMORY)                      \
            && (char *)(MEMORY) < (HYBRID_ALLOCATOR)->end)

/* Given an address (MEMORY) inside a memory block currently allocated by
 * hybrid allocator that uses (ARRAY) as its stack, expands to a nonzero
 * expression if and only if (MEMORY) was allocated inside (ARRAY),
 * as opposed to having been allocated by the delegate allocator.
 *
 * See also HYBRID_ALLOCATOR_STACK_CONTAINS, the alternative
 * for when (ARRAY) is not currently in scope.
 *
 * WARNING: This macro evaluates its arguments more than once.
 */
#define STACK_FOR_HYBRID_ALLOCATOR_CONTAINS(ARRAY, MEMORY)              \
    ((char *)(ARRAY) <= (char *)(MEMORY)                                \
            && (char *)(MEMORY) < (char *)VRPSX_ARRAY_END(ARRAY))


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* HYBRID_ALLOCATOR_H */
