/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * This header defines utility macros and types that may assist
 * interface usage, but for compatibility reasons should never be
 * included in headers that define interfaces.
 *
 * In particular, this header should never be construed as contributing
 * to the definition of any interface--not even to its documentation.
 * Interfaces should be defined without reference to this header.
 *
 * Rather, this header is intended to make it easier to use the currently
 * supported interfaces, and in future we may need to modify or replace this
 * header in order to support a new interface.  That future change may be
 * source-incompatible, but so long as no interface definition makes reference
 * to this header or to its utilities, binary compatibility will be preserved.
 *
 * Furthermore, when used as intended, none of the features of this header
 * will affect linkage, and therefore you may link together compilations
 * of translation units that include different versions of this header.
 *
 * Pointer Utilities:
 *
 * Some of the macros defined in this header are intended to make it easier
 * to follow a particular idiom of interface control, such as single-owner
 * or reference-counted, by binding the relevant owner concept to pointer
 * variables using the RAII principle (though without language support):
 *
 *     http://en.wikipedia.org/wiki/Resource_Acquisition_Is_Initialization)
 *
 * As with other macros in this header, these pointer utilities should not be
 * construed as affecting the definition of any interface--interfaces should
 * specify the underlying idioms, NOT these macros.  Thus if you prefer you
 * can completely ignore these utilities and instead directly manipulate
 * interfaces, with or without use of RAII principles as such.
 *
 * Macros that Expand to Blocks of Code:
 *
 * Until C99 becomes more widespread, we use macros that expand to blocks of
 * code instead of inline functions.  To make variable name collisions less
 * probable, we append the hexadecimal representations of random integers
 * to the names of local variables declared by such expansions.  The random
 * choice can be shared when a collision is not possible, just for convenience
 * when comparing otherwise-similar macros with "diff", but when in doubt
 * please make a new random choice to keep the variable names distinct.
 * (Of course, with inline functions there is no need for such measures.)
 */

#ifndef VRPSX_UTILITY_H
#define VRPSX_UTILITY_H 1

#include "vrpsx_preprocessor.h"
#include "../interface/vrpsx_arch.h"
#include "../interface/vrpsx_stdint.h"
#include <stddef.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */

#if defined(VRPSX_MSC)
#include <windows.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */
#endif /* compiler */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* Given an expression of array type, expands to a constant expression whose
 * value is the dimension of that array.  (For multidimensional arrays this
 * macro returns the outermost dimension.)
 */
#define VRPSX_ARRAY_DIMENSION(ARRAY)  (sizeof (ARRAY) / sizeof (ARRAY)[0])

/* ARRAY + VRPSX_ARRAY_DIMENSION(ARRAY)
 */
#define VRPSX_ARRAY_END(ARRAY)  (ARRAY + VRPSX_ARRAY_DIMENSION(ARRAY))


/* The alignment of every C data type divides the alignment of this data type.
 *
 * (If you find an exception to this rule, then please expand the union.)
 */
typedef union vrpsx_maximally_aligned_union {
    unsigned long long _unsigned_long_long;
    long double _long_double;
    void *_void_ptr;
    void (*_function_ptr)(void);
    size_t _size_t;
    vrpsx_uint64_t _uint64_t;
} vrpsx_maximally_aligned_t;

/* An implementation detail of VRPSX_MAXIMUM_ALIGNMENT.
 */
typedef struct vrpsx_maximum_alignment_helper_struct {
    char _minimally_aligned;
    vrpsx_maximally_aligned_t _maximally_aligned;
} vrpsx_maximum_alignment_helper_t;

/* A multiple of the maximum alignment required for any C data type.
 */
#define VRPSX_MAXIMUM_ALIGNMENT                                         \
    (offsetof(vrpsx_maximum_alignment_helper_t, _maximally_aligned))


/* The type to use for reference counts.  For thread-safety,
 * you must use atomic increment and decrement functions.
 */
#if VRPSX_ATOMIC_BITS == 64
typedef vrpsx_int64_t vrpsx_reference_count_t;
#elif VRPSX_ATOMIC_BITS == 32
typedef vrpsx_int32_t vrpsx_reference_count_t;
#else /* VRPSX_ATOMIC_BITS */
#error unsupported atomic word size
#endif /* VRPSX_ATOMIC_BITS */


/* Reference Count Utilities:
 *
 * VRPSX_REFERENCE_COUNT_INCR(COUNT)
 * VRPSX_REFERENCE_COUNT_DECR(COUNT)
 *
 *    Atomically increments/decrements *COUNT.
 *
 *    Evaluates COUNT exactly once.
 *
 * VRPSX_REFERENCE_COUNT_INCR_READ(COUNT, NEW_COUNT)
 * VRPSX_REFERENCE_COUNT_DECR_READ(COUNT, NEW_COUNT)
 *
 *    Atomically increments/decrements *COUNT.  As a possibly separate
 *    operation that need not be synchronized, writes to *NEW_COUNT the
 *    new value of *COUNT that was written during the atomic operation.
 *
 *    Evaluates the pointer values COUNT and NEW_COUNT exactly once.
 *    It is assumed that COUNT != NEW_COUNT (as pointer values).
 *
 *    Typically NEW_COUNT should be on the stack, thread-specific,
 *    or have access to it synchronized in some fashion.
 */

#if defined(VRPSX_GNUC)

#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 1)

/* The __sync... functions are first documented in GCC 4.1. */

#if VRPSX_ATOMIC_BITS == 64

#define VRPSX_REFERENCE_COUNT_INCR(COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);    \
    __sync_add_and_fetch(count_d9d169ef, 1);    \
}

#define VRPSX_REFERENCE_COUNT_DECR(COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);    \
    __sync_sub_and_fetch(count_d9d169ef, 1);    \
}

#define VRPSX_REFERENCE_COUNT_INCR_READ(COUNT, NEW_COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);                    \
    *(NEW_COUNT) = __sync_add_and_fetch(count_d9d169ef, 1);     \
}

#define VRPSX_REFERENCE_COUNT_DECR_READ(COUNT, NEW_COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);                    \
    *(NEW_COUNT) = __sync_sub_and_fetch(count_d9d169ef, 1);     \
}

#elif VRPSX_ATOMIC_BITS == 32

#define VRPSX_REFERENCE_COUNT_INCR(COUNT)  {    \
    vrpsx_int32_t *count_d9d169ef = (COUNT);    \
    __sync_add_and_fetch(count_d9d169ef, 1);    \
}

#define VRPSX_REFERENCE_COUNT_DECR(COUNT)  {    \
    vrpsx_int32_t *count_d9d169ef = (COUNT);    \
    __sync_sub_and_fetch(count_d9d169ef, 1);    \
}

#define VRPSX_REFERENCE_COUNT_INCR_READ(COUNT, NEW_COUNT)  {    \
    vrpsx_int32_t *count_d9d169ef = (COUNT);                    \
    *(NEW_COUNT) = __sync_add_and_fetch(count_d9d169ef, 1);     \
}

#define VRPSX_REFERENCE_COUNT_DECR_READ(COUNT, NEW_COUNT)  {    \
    vrpsx_int32_t *count_d9d169ef = (COUNT);                    \
    *(NEW_COUNT) = __sync_sub_and_fetch(count_d9d169ef, 1);     \
}

#endif /* VRPSX_ATOMIC_BITS */

#else /* GCC earlier than 4.1.0 */

/* The __sync... functions might not be available in this GCC version;
 * instead fall back to inline assembler on supported architecture(s):
 */

#if defined(VRPSX_X86)

#if VRPSX_ATOMIC_BITS == 64

#define VRPSX_REFERENCE_COUNT_INCR(COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);    \
                                                \
    asm volatile(                               \
        "lock xadd %0, %1"                      \
        :                                       \
        : "r" (1L), "m" (*count_d9d169ef)       \
        : "memory" );                           \
}

#define VRPSX_REFERENCE_COUNT_DECR(COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);    \
                                                \
    asm volatile(                               \
        "lock xadd %0, %1"                      \
        :                                       \
        : "r" (-1L), "m" (*count_d9d169ef)      \
        : "memory" );                           \
}

#define VRPSX_REFERENCE_COUNT_INCR_READ(COUNT, NEW_COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);                    \
    vrpsx_int64_t old_d9d169ef;                                 \
                                                                \
    asm volatile(                                               \
        "lock xadd %0, %2"                                      \
        :"=r" (old_d9d169ef)                                    \
        : "0" (1L), "m" (*count_d9d169ef)                       \
        : "memory" );                                           \
                                                                \
    *(NEW_COUNT) = (old_d9d169ef + 1);                          \
}

#define VRPSX_REFERENCE_COUNT_DECR_READ(COUNT, NEW_COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);                    \
    vrpsx_int64_t old_d9d169ef;                                 \
                                                                \
    asm volatile(                                               \
        "lock xadd %0, %2"                                      \
        :"=r" (old_d9d169ef)                                    \
        : "0" (-1L), "m" (*count_d9d169ef)                      \
        : "memory" );                                           \
                                                                \
    *(NEW_COUNT) = (old_d9d169ef - 1);                          \
}

#elif VRPSX_ATOMIC_BITS == 32

#define VRPSX_REFERENCE_COUNT_INCR(COUNT)  {    \
    vrpsx_int32_t *count_d9d169ef = (COUNT);    \
                                                \
    asm volatile(                               \
        "lock xadd %0, %1"                      \
        :                                       \
        : "r" (1), "m" (*count_d9d169ef)        \
        : "memory" );                           \
}

#define VRPSX_REFERENCE_COUNT_DECR(COUNT)  {    \
    vrpsx_int32_t *count_d9d169ef = (COUNT);    \
                                                \
    asm volatile(                               \
        "lock xadd %0, %1"                      \
        :                                       \
        : "r" (-1), "m" (*count_d9d169ef)       \
        : "memory" );                           \
}

#define VRPSX_REFERENCE_COUNT_INCR_READ(COUNT, NEW_COUNT)  {    \
    vrpsx_int32_t *count_d9d169ef = (COUNT);                    \
    vrpsx_int32_t old_d9d169ef;                                 \
                                                                \
    asm volatile(                                               \
        "lock xadd %0, %2"                                      \
        :"=r" (old_d9d169ef)                                    \
        : "0" (1), "m" (*count_d9d169ef)                        \
        : "memory" );                                           \
                                                                \
    *(NEW_COUNT) = (old_d9d169ef + 1);                          \
}

#define VRPSX_REFERENCE_COUNT_DECR_READ(COUNT, NEW_COUNT)  {    \
    vrpsx_int32_t *count_d9d169ef = (COUNT);                    \
    vrpsx_int32_t old_d9d169ef;                                 \
                                                                \
    asm volatile(                                               \
        "lock xadd %0, %2"                                      \
        :"=r" (old_d9d169ef)                                    \
        : "0" (-1), "m" (*count_d9d169ef)                       \
        : "memory" );                                           \
                                                                \
    *(NEW_COUNT) = (old_d9d169ef - 1);                          \
}

#endif /* VRPSX_ATOMIC_BITS */

#else /* instruction set */

#error upgrade GCC or write inline assembly code for this architecture

#endif /* instruction set */

#endif /* GCC version */

#elif defined(VRPSX_MSC)

#if VRPSX_ATOMIC_BITS == 64

#define VRPSX_REFERENCE_COUNT_INCR(COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);    \
    InterlockedIncrement64(count_d9d169ef);     \
}

#define VRPSX_REFERENCE_COUNT_DECR(COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);    \
    InterlockedDecrement64(count_d9d169ef);     \
}

#define VRPSX_REFERENCE_COUNT_INCR_READ(COUNT, NEW_COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);                    \
    *(NEW_COUNT) = InterlockedIncrement64(count_d9d169ef));     \
}

#define VRPSX_REFERENCE_COUNT_DECR_READ(COUNT, NEW_COUNT)  {    \
    vrpsx_int64_t *count_d9d169ef = (COUNT);                    \
    *(NEW_COUNT) = InterlockedDecrement64(count_d9d169ef);      \
}

#elif VRPSX_ATOMIC_BITS == 32

/* On 32-bit Windows InterlockedIncrement and InterlockedDecrement
 * have argument type (LONG volatile *), which is a 32-bit signed
 * integer type but might nevertheless be distinct from the 32-bit
 * signed integer type we chose for vrpsx_reference_count_t.
 * We cast through this cosmetic distinction, which has no effect
 * on the binary compatibility of reference-counted interfaces.
 */

#define VRPSX_REFERENCE_COUNT_INCR(COUNT)  {                    \
    vrpsx_int32_t *count_d9d169ef = (COUNT);                    \
    InterlockedIncrement((LONG volatile *)count_d9d169ef);      \
}

#define VRPSX_REFERENCE_COUNT_DECR(COUNT)  {                    \
    vrpsx_int32_t *count_d9d169ef = (COUNT);                    \
    InterlockedDecrement((LONG volatile *)count_d9d169ef);      \
}

#define VRPSX_REFERENCE_COUNT_INCR_READ(COUNT, NEW_COUNT)  {            \
    vrpsx_int32_t *count_d9d169ef = (COUNT);                            \
    *(NEW_COUNT) = InterlockedIncrement((LONG volatile *)count_d9d169ef); \
}

#define VRPSX_REFERENCE_COUNT_DECR_READ(COUNT, NEW_COUNT)  {            \
    vrpsx_int32_t *count_d9d169ef = (COUNT);                            \
    *(NEW_COUNT) = InterlockedDecrement((LONG volatile *)count_d9d169ef); \
}

#endif /* VRPSX_ATOMIC_BITS */

#endif /* compiler */


/* Expands to a constant provided by an interface.
 *
 * This macro is intended for convenience only; you can
 * use the virtual function table directly if you prefer.
 *
 * The INTERFACE argument is evaluated exactly once.
 * It must not be NULL.
 */
#define VRPSX_CONSTANT(INTERFACE, CONSTANT)     \
    ((INTERFACE)->vtbl->CONSTANT)


/* Calls the named interface operation, which in C++ terms would be
 * a virtual member function, but in C is a function pointer member
 * of the virtual function table structure pointed to by the "vtbl"
 * member of the interface.
 *
 * This macro is intended for convenience only; you can
 * use the virtual function table directly if you prefer.
 *
 * This particular macro handles the case in which
 * the only argument is the interface itself.
 *
 * NOTE: The INTERFACE argument may be evaluated more than once.
 * It must not be NULL.
 *
 * See also: VRPSX_DELETE0
 */
#define VRPSX_CALL0(INTERFACE, OPERATION)       \
    ((INTERFACE)->vtbl->OPERATION((INTERFACE)))

/* Like VRPSX_CALL0, but for operations that take 1 argument in
 * addition to the interface.  Unlike INTERFACE, that additional
 * argument is evaluated exactly once.
 */
#define VRPSX_CALL1(INTERFACE, OPERATION, A1)   \
    ((INTERFACE)->vtbl->OPERATION((INTERFACE),  \
         (A1)))

/* Like VRPSX_CALL0, but for operations that take 2 arguments in
 * addition to the interface.  Unlike INTERFACE, those additional
 * arguments are each evaluated exactly once.
 */
#define VRPSX_CALL2(INTERFACE, OPERATION, A1, A2)       \
    ((INTERFACE)->vtbl->OPERATION((INTERFACE),          \
         (A1), (A2)))

/* Like VRPSX_CALL0, but for operations that take 3 arguments in
 * addition to the interface.  Unlike INTERFACE, those additional
 * arguments are each evaluated exactly once.
 */
#define VRPSX_CALL3(INTERFACE, OPERATION, A1, A2, A3)   \
    ((INTERFACE)->vtbl->OPERATION((INTERFACE),          \
         (A1), (A2), (A3)))

/* Like VRPSX_CALL0, but for operations that take 4 arguments in
 * addition to the interface.  Unlike INTERFACE, those additional
 * arguments are each evaluated exactly once.
 */
#define VRPSX_CALL4(INTERFACE, OPERATION, A1, A2, A3, A4)       \
    ((INTERFACE)->vtbl->OPERATION((INTERFACE),                  \
         (A1), (A2), (A3), (A4)))

/* Like VRPSX_CALL0, but for operations that take 5 arguments in
 * addition to the interface.  Unlike INTERFACE, those additional
 * arguments are each evaluated exactly once.
 */
#define VRPSX_CALL5(INTERFACE, OPERATION, A1, A2, A3, A4, A5)   \
    ((INTERFACE)->vtbl->OPERATION((INTERFACE),                  \
         (A1), (A2), (A3), (A4), (A5)))

/* Like VRPSX_CALL0, but for operations that take 6 arguments in
 * addition to the interface.  Unlike INTERFACE, those additional
 * arguments are each evaluated exactly once.
 */
#define VRPSX_CALL6(INTERFACE, OPERATION, A1, A2, A3, A4, A5, A6)       \
    ((INTERFACE)->vtbl->OPERATION((INTERFACE),                          \
         (A1), (A2), (A3), (A4), (A5), (A6)))

/* Like VRPSX_CALL0, but for operations that take 7 arguments in
 * addition to the interface.  Unlike INTERFACE, those additional
 * arguments are each evaluated exactly once.
 */
#define VRPSX_CALL7(INTERFACE, OPERATION, A1, A2, A3, A4, A5, A6, A7)   \
    ((INTERFACE)->vtbl->OPERATION((INTERFACE),                          \
         (A1), (A2), (A3), (A4), (A5), (A6), (A7)))


/* Like VRPSX_CALL0 but for operations whose first
 * argument is a pointer to a pointer to the interface.
 */
#define VRPSX_DISPOSE0(INTERFACE, OPERATION)            \
    ((*(INTERFACE))->vtbl->OPERATION((INTERFACE)))

/* Like VRPSX_CALL1 but for operations whose first
 * argument is a pointer to a pointer to the interface.
 */
#define VRPSX_DISPOSE1(INTERFACE, OPERATION, A1)        \
    ((*(INTERFACE))->vtbl->OPERATION((INTERFACE),       \
         (A1)))

/* Like VRPSX_CALL2 but for operations whose first
 * argument is a pointer to a pointer to the interface.
 */
#define VRPSX_DISPOSE2(INTERFACE, OPERATION, A1, A2)    \
    ((*(INTERFACE))->vtbl->OPERATION((INTERFACE),       \
         (A1), (A2)))

/* Like VRPSX_CALL3 but for operations whose first
 * argument is a pointer to a pointer to the interface.
 */
#define VRPSX_DISPOSE3(INTERFACE, OPERATION, A1, A2, A3)        \
    ((*(INTERFACE))->vtbl->OPERATION((INTERFACE),               \
         (A1), (A2), (A3)))

/* Like VRPSX_CALL4 but for operations whose first
 * argument is a pointer to a pointer to the interface.
 */
#define VRPSX_DISPOSE4(INTERFACE, OPERATION, A1, A2, A3, A4)    \
    ((*(INTERFACE))->vtbl->OPERATION((INTERFACE),               \
         (A1), (A2), (A3), (A4)))

/* Like VRPSX_CALL5 but for operations whose first
 * argument is a pointer to a pointer to the interface.
 */
#define VRPSX_DISPOSE5(INTERFACE, OPERATION, A1, A2, A3, A4, A5)        \
    ((*(INTERFACE))->vtbl->OPERATION((INTERFACE),                       \
         (A1), (A2), (A3), (A4), (A5)))

/* Like VRPSX_CALL6 but for operations whose first
 * argument is a pointer to a pointer to the interface.
 */
#define VRPSX_DISPOSE6(INTERFACE, OPERATION, A1, A2, A3, A4, A5, A6)    \
    ((*(INTERFACE))->vtbl->OPERATION((INTERFACE),                       \
         (A1), (A2), (A3), (A4), (A5), (A6)))

/* Like VRPSX_CALL7 but for operations whose first
 * argument is a pointer to a pointer to the interface.
 */
#define VRPSX_DISPOSE7(INTERFACE, OPERATION, A1, A2, A3, A4, A5, A6, A7) \
    ((*(INTERFACE))->vtbl->OPERATION((INTERFACE),                       \
         (A1), (A2), (A3), (A4), (A5), (A6), (A7)))


/* Static check that &IMPLEMENTATION->MEMBER has the same type as INTERFACE.
 * Expands to an expression with the same value as IMPLEMENTATION, which
 * is evaluated exactly once (though it appears multiple times in the macro).
 * The INTERFACE argument is never evaluated, though it is used in the check.
 */
#define VRPSX_UPCAST_CHECK(IMPLEMENTATION, MEMBER, INTERFACE)           \
    ( (IMPLEMENTATION)                                                  \
      + ( sizeof ((char)(&(IMPLEMENTATION) -> MEMBER - (INTERFACE)))    \
          - sizeof(char) ) )

/* Assuming an instance of IMPLEMENTATION has a member named MEMBER at
 * non-NULL address INTERFACE, returns the address of that instance.
 * INTERFACE is evaluated exactly once.
 *
 * Warning: you may quietly lose "const" qualification.
 */
#define VRPSX_UPCAST_NONNULL(IMPLEMENTATION, MEMBER, INTERFACE)         \
    (VRPSX_UPCAST_CHECK(                                                \
        (IMPLEMENTATION *)(                                             \
            (char*)(INTERFACE) - offsetof(IMPLEMENTATION, MEMBER)),     \
        MEMBER, (INTERFACE)))

/* Like VRPSX_UPCAST_NONNULL but declares a variable in which
 * to store the result of the cast.  The terminating semicolon
 * is omitted to help smart editors.  Typically used in methods.
 *
 * Warning: you may quietly lose "const" qualification.
 * To compensate, be sure to pass the correct QUALIFIER.
 */
#define VRPSX_UPCAST_NONNULL_DECL(IMPLEMENTATION, QUALIFIER, NAME, MEMBER, \
                                  INTERFACE)                            \
    IMPLEMENTATION QUALIFIER * NAME                                     \
        = VRPSX_UPCAST_NONNULL(IMPLEMENTATION, MEMBER, (INTERFACE))

/* Like VRPSX_UPCAST_NONNULL(IMPLEMENTATION, MEMBER, INTERFACE),
 * but also transforms NULL to NULL, and in order to avoid evaluating
 * INTERFACE twice for that additional logic, this macro expands to
 * a block that assigns to *OUTPUT, instead of to an expression.
 * The expression OUTPUT is evaluated exactly once.
 *
 * Warning: you may quietly lose "const" qualification.
 */
#define VRPSX_UPCAST(OUTPUT, IMPLEMENTATION, MEMBER, INTERFACE)  {      \
    char *interface_382abbfb = (char*)(INTERFACE);                      \
    *(OUTPUT) = VRPSX_UPCAST_CHECK(                                     \
        ( interface_382abbfb                                            \
            ? (IMPLEMENTATION *)(                                       \
                (char*)interface_382abbfb - offsetof(IMPLEMENTATION, MEMBER)) \
            : (IMPLEMENTATION *)interface_382abbfb ),                   \
        MEMBER, INTERFACE);                                             \
}


/* Initializes an interface pointer according to the single-owner idiom:
 * sets *TARGET to *SOURCE and clears *SOURCE.
 *
 * Evaluates TARGET and SOURCE exactly once.
 * If TARGET == SOURCE then that one pointer variable keeps its value.
 */
#define VRPSX_SINGLE_OWNER_MOVE(TYPE, TARGET, SOURCE)  {        \
    TYPE **target_24e93942 = (TARGET);                          \
    TYPE **source_24e93942 = (SOURCE);                          \
    TYPE *new_24e93942 = *source_24e93942;                      \
                                                                \
    *source_24e93942 = NULL;                                    \
    *target_24e93942 = new_24e93942;                            \
}

/* Clears an interface pointer according to the single-owner idiom.
 * The interface must have a typical "disown" operation.
 *
 * Evaluates TARGET exactly once.
 */
#define VRPSX_SINGLE_OWNER_CLEAR(TYPE, TARGET)  {       \
    TYPE **target_24e93942 = (TARGET);                  \
                                                        \
    if (*target_24e93942) {                             \
        VRPSX_DISPOSE0(target_24e93942, disown);        \
    }                                                   \
}

/* Like VRPSX_SINGLE_OWNER_CLEAR, but assumes that the "disown"
 * operation returns a value, and stores that value into *RESULT.
 *
 * NOTE: If *TARGET was already NULL, then *RESULT keeps its old value.
 *
 * Evaluates TARGET exactly once and RESULT at most once.
 * RESULT is evaluated if and only if *TARGET was not already NULL.
 */
#define VRPSX_SINGLE_OWNER_CLOSE(TYPE, TARGET, RESULT)  {       \
    TYPE **target_24e93942 = (TARGET);                          \
                                                                \
    if (*target_24e93942) {                                     \
        *(RESULT) = VRPSX_DISPOSE0(target_24e93942, disown);    \
    }                                                           \
}

/* Like VRPSX_SINGLE_OWNER_CLOSE, but for when
 * there is a "System Reference On Error" argument.
 *
 * Evaluates SROE at most once.
 * SROE is evaluated if and only if *TARGET was not already NULL.
 */
#define VRPSX_SINGLE_OWNER_CLOSE_SROE(TYPE, TARGET, RESULT, SROE)  {    \
    TYPE **target_24e93942 = (TARGET);                                  \
                                                                        \
    if (*target_24e93942) {                                             \
        *(RESULT) = VRPSX_DISPOSE1(target_24e93942, disown, (SROE));    \
    }                                                                   \
}

/* Assigns an interface pointer according to the single-owner idiom:
 * sets *TARGET to *SOURCE, clears *SOURCE, and if the previous value of
 * *TARGET was not NULL, then invokes the "disown" operation on that old value.
 *
 * Evaluates TARGET and SOURCE exactly once.
 * If TARGET == SOURCE then that one pointer variable keeps its value.
 */
#define VRPSX_SINGLE_OWNER_DISPLACE(TYPE, TARGET, SOURCE)  {            \
    TYPE **target_24e93942 = (TARGET);                                  \
    TYPE **source_24e93942 = (SOURCE);                                  \
    TYPE *new_24e93942 = *source_24e93942;                              \
    TYPE *old_24e93942;                                                 \
                                                                        \
    /* Clear source first in case target and source are same. */        \
    *source_24e93942 = NULL;                                            \
                                                                        \
    old_24e93942 = *target_24e93942;                                    \
    *target_24e93942 = new_24e93942;                                    \
                                                                        \
    if (old_24e93942) {                                                 \
        VRPSX_DISPOSE0(&old_24e93942, disown);                          \
    }                                                                   \
}

/* Like VRPSX_SINGLE_OWNER_CLEAR except that the pointer
 * to be cleared is a pointer to an implementation whose
 * interface is in the specified member.
 */
#define VRPSX_SINGLE_OWNER_CLEAR_IMPLEMENTATION(                        \
    IMPLEMENTATION, MEMBER, INTERFACE, TARGET)  {                       \
    IMPLEMENTATION **target_24e93942 = (TARGET);                        \
                                                                        \
    if (*target_24e93942) {                                             \
        INTERFACE *interface_24e93942 = &(*target_24e93942)->MEMBER;    \
        *target_24e93942 = NULL;                                        \
        VRPSX_DISPOSE0(&interface_24e93942, disown);                    \
    }                                                                   \
}

/* Like VRPSX_SINGLE_OWNER_CLOSE except that the pointer
 * to be cleared is a pointer to an implementation whose
 * interface is in the specified member.
 */
#define VRPSX_SINGLE_OWNER_CLOSE_IMPLEMENTATION(                        \
    IMPLEMENTATION, MEMBER, INTERFACE, TARGET, RESULT)  {               \
    IMPLEMENTATION **target_24e93942 = (TARGET);                        \
                                                                        \
    if (*target_24e93942) {                                             \
        INTERFACE *interface_24e93942 = &(*target_24e93942)->MEMBER;    \
        *target_24e93942 = NULL;                                        \
        *(RESULT) = VRPSX_DISPOSE0(&interface_24e93942, disown);        \
    }                                                                   \
}


/* Initializes an interface pointer according to a modified single-owner
 * idiom in which interface deletion may give rise to another interface:
 * sets *TARGET to *SOURCE and clears *SOURCE.
 *
 * Evaluates TARGET and SOURCE exactly once.
 * If TARGET == SOURCE then that one pointer variable keeps its value.
 */
#define VRPSX_SINGLE_OWNER_LIST_MOVE(TYPE, TARGET, SOURCE)      \
    VRPSX_SINGLE_OWNER_MOVE(TYPE, TARGET, SOURCE)

/* Clears an interface pointer according to a modified single-owner
 * idiom in which interface deletion may give rise to another interface.
 * The interface pointer *TARGET is set to NULL.
 *
 * Evaluates TARGET exactly once.
 */
#define VRPSX_SINGLE_OWNER_LIST_CLEAR(TYPE, TARGET)  {  \
    TYPE **target_24e93942 = (TARGET);                  \
                                                        \
    if (*target_24e93942) {                             \
        VRPSX_DISPOSE1(target_24e93942, next, NULL);    \
    }                                                   \
}

/* Like VRPSX_SINGLE_OWNER_LIST_CLEAR, but for when
 * there is a "System Reference On Error" argument.
 */
#define VRPSX_SINGLE_OWNER_LIST_CLEAR_SROE(TYPE, TARGET)  {     \
    TYPE **target_24e93942 = (TARGET);                          \
                                                                \
    if (*target_24e93942) {                                     \
        VRPSX_DISPOSE2(target_24e93942, next, NULL, NULL);      \
    }                                                           \
}

/* Replaces an interface with the next interface, if any, according
 * to a modified single-owner idiom in which interface deletion may
 * give rise to another interface.
 *
 * Evaluates TARGET and NEXT exactly once.
 * Works even if TARGET == NEXT (as pointer values); in that case
 * *TARGET will hold the next interface, or NULL if there is none.
 */
#define VRPSX_SINGLE_OWNER_LIST_NEXT(TYPE, TARGET, NEXT)  {     \
    TYPE **target_24e93942 = (TARGET);                          \
                                                                \
    VRPSX_DISPOSE1(target_24e93942, next, (NEXT));              \
}

/* Like VRPSX_SINGLE_OWNER_LIST_NEXT but stores the value returned by
 * the "next" operation into *RESULT.  RESULT is evaluated exactly once.
 */
#define VRPSX_SINGLE_OWNER_LIST_NEXT_CHECK(TYPE, TARGET, NEXT, RESULT)  { \
    TYPE **target_24e93942 = (TARGET);                                  \
                                                                        \
    *(RESULT) = VRPSX_DISPOSE1(target_24e93942, next, (NEXT));          \
}

/* Like VRPSX_SINGLE_OWNER_LIST_NEXT_CHECK, but for when
 * there is a "System Reference On Error" argument.
 *
 * Evaluates SROE at most once.
 * SROE is evaluated if and only if *TARGET was not already NULL.
 */
#define VRPSX_SINGLE_OWNER_LIST_NEXT_CHECK_SROE(TYPE, TARGET, NEXT, RESULT, SROE) {\
    TYPE **target_24e93942 = (TARGET);                                  \
                                                                        \
    *(RESULT) = VRPSX_DISPOSE2(target_24e93942, next, (NEXT), (SROE));  \
}

/* Assigns an interface pointer according to a modified single-owner
 * idiom in which interface deletion may give rise to another interface:
 * sets *TARGET to *SOURCE, clears *SOURCE, and if the previous value of
 * *TARGET was not NULL, then invokes the "next" operation on that old value
 * with a NULL second argument to indicate we do not want the next interface.
 *
 * Evaluates TARGET and SOURCE exactly once.
 * If TARGET == SOURCE then that one pointer variable keeps its value.
 */
#define VRPSX_SINGLE_OWNER_LIST_DISPLACE(TYPE, TARGET, SOURCE)  {       \
    TYPE **target_24e93942 = (TARGET);                                  \
    TYPE **source_24e93942 = (SOURCE);                                  \
    TYPE *new_24e93942 = *source_24e93942;                              \
    TYPE *old_24e93942;                                                 \
                                                                        \
    /* Clear source first in case target and source are same. */        \
    *source_24e93942 = NULL;                                            \
                                                                        \
    old_24e93942 = *target_24e93942;                                    \
    *target_24e93942 = new_24e93942;                                    \
                                                                        \
    if (old_24e93942) {                                                 \
        VRPSX_DISPOSE1(&old_24e93942, next, NULL);                      \
    }                                                                   \
}

/* Like VRPSX_SINGLE_OWNER_LIST_DISPLACE, but for when
 * there is a "System Reference On Error" argument.
 */
#define VRPSX_SINGLE_OWNER_LIST_DISPLACE_SROE(TYPE, TARGET, SOURCE)  {  \
    TYPE **target_24e93942 = (TARGET);                                  \
    TYPE **source_24e93942 = (SOURCE);                                  \
    TYPE *new_24e93942 = *source_24e93942;                              \
    TYPE *old_24e93942;                                                 \
                                                                        \
    /* Clear source first in case target and source are same. */        \
    *source_24e93942 = NULL;                                            \
                                                                        \
    old_24e93942 = *target_24e93942;                                    \
    *target_24e93942 = new_24e93942;                                    \
                                                                        \
    if (old_24e93942) {                                                 \
        VRPSX_DISPOSE2(&old_24e93942, next, NULL, NULL);                \
    }                                                                   \
}


/* Initializes a pointer to a reference-counted interface:
 * sets *TARGET to POINTER, and if POINTER != NULL then
 * calls the "acquire" operation on POINTER.
 *
 * Note that if POINTER will always be NULL, then it is simpler
 * to directly initialize *TARGET than to expand this macro.
 *
 * Evaluates TARGET and POINTER exactly once.
 * (POINTER is read before *TARGET is written, meaning that you could
 * turn a non-owning pointer into an owning one if POINTER is *TARGET,
 * but such usage is rather confusing and so we recommend against it.)
 */
#define VRPSX_COUNTED_REFERENCE_ACQUIRE(TYPE, TARGET, POINTER)  {       \
    TYPE **target_24e93942 = (TARGET);                                  \
    TYPE *pointer_24e93942 = (POINTER);                                 \
                                                                        \
    *target_24e93942 = pointer_24e93942;                                \
                                                                        \
    if (pointer_24e93942) {                                             \
        VRPSX_CALL0(pointer_24e93942, acquire);                         \
    }                                                                   \
}

/* Initializes a pointer to a reference-counted interface
 * by stealing a reference from another pointer variable:
 * sets *TARGET to *SOURCE and clears *SOURCE.
 *
 * Evaluates TARGET and SOURCE exactly once.
 * If TARGET == SOURCE then that one pointer variable keeps its value.
 */
#define VRPSX_COUNTED_REFERENCE_MOVE(TYPE, TARGET, SOURCE)      \
    VRPSX_SINGLE_OWNER_MOVE(TYPE, TARGET, SOURCE)

/* Clears a pointer to a reference-counted interface:
 * clears *TARGET and releases any previous reference
 * by calling its "release" operation.
 *
 * Evaluates TARGET exactly once.
 */
#define VRPSX_COUNTED_REFERENCE_CLEAR(TYPE, TARGET)  {  \
    TYPE **target_24e93942 = (TARGET);                  \
                                                        \
    if (*target_24e93942) {                             \
        VRPSX_DISPOSE0(target_24e93942, release);       \
    }                                                   \
}

/* Assigns a pointer to a reference-counted interface:
 * sets *TARGET to POINTER, calling "acquire" on POINTER if it is not NULL,
 * then call the "release" operation on any previous reference in *TARGET.
 *
 * The "acquire" calls always come before the "release" calls in order
 * to avoid premature disposal when reassigning the same pointer value.
 *
 * Evaluates TARGET and POINTER exactly once.
 * POINTER is read before *TARGET is written, and therefore
 * VRPSX_COUNTED_REFERENCE_ASSIGN(TYPE, &var, var) has no net effect.
 */
#define VRPSX_COUNTED_REFERENCE_ASSIGN(TYPE, TARGET, POINTER)  {        \
    TYPE **target_24e93942 = (TARGET);                                  \
    TYPE *new_24e93942 = (POINTER);                                     \
    TYPE *old_24e93942 = *target_24e93942;                              \
                                                                        \
    *target_24e93942 = new_24e93942;                                    \
                                                                        \
    if (new_24e93942) {                                                 \
        VRPSX_CALL0(new_24e93942, acquire);                             \
    }                                                                   \
                                                                        \
    if (old_24e93942) {                                                 \
        VRPSX_DISPOSE0(&old_24e93942, release);                         \
    }                                                                   \
}

/* Assigns to a pointer to a reference-counted interface
 * by stealing a reference from another pointer variable:
 * sets *TARGET to *SOURCE, clears *SOURCE, and calls the
 * "release" operation on any previous reference in *TARGET.
 *
 * Evaluates TARGET and SOURCE exactly once.
 * If TARGET == SOURCE then that one pointer variable keeps its value.
 */
#define VRPSX_COUNTED_REFERENCE_DISPLACE(TYPE, TARGET, SOURCE)  {       \
    TYPE **target_24e93942 = (TARGET);                                  \
    TYPE **source_24e93942 = (SOURCE);                                  \
    TYPE *new_24e93942 = *source_24e93942;                              \
    TYPE *old_24e93942;                                                 \
                                                                        \
    /* Clear source first in case target and source are same. */        \
    *source_24e93942 = NULL;                                            \
                                                                        \
    *old_24e93942 = *target_24e93942;                                   \
    *target_24e93942 = new_24e93942;                                    \
                                                                        \
    if (old_24e93942) {                                                 \
        VRPSX_DISPOSE0(&old_24e93942, release);                         \
    }                                                                   \
}


/* The header <vrpsx_identified_function.h> defines situations in which
 * you might define a C identifier consisting of "_" followed by the
 * identifier for the interface defined by <vrpsx_identified_function.h>.
 *
 * After you include
 *
 *     <vrpsx_identified_function_identifier.h>
 *
 * this macro will expand to the C identifier for that symbol.
 * (But the aforementioned interfaces do not even mention this macro,
 * much less require you to use it.  It is merely a convenience.)
 *
 * NOTE: If you want the equivalent string literal, then please
 * use VRPSX_IDENTIFIED_FUNCTION_LIST_STR instead of stringizing
 * this macro, because the latter approach would compile in even
 * if you forgot to include a definition of the identifier.
 */
#define VRPSX_IDENTIFIED_FUNCTION_LIST                  \
    vrpsx_identified_function_IDENTIFIER_CAT(_,)

/* Like VRPSX_IDENTIFIED_FUNCTION_LIST but expands to a string literal.
 */
#define VRPSX_IDENTIFIED_FUNCTION_LIST_STR              \
    "_" vrpsx_identified_function_IDENTIFIER_STR

/* Expands to PREFIX_abcdef01 but with abcdef01 replaced
 * by the value of the macro INTERFACE_IDENTIFIER_RAW.
 *
 * Assuming that you are following typical naming practices,
 * you can use this macro with VRPSX_DECLARE_IDENTIFIED_FUNCTION
 * and VRPSX_MENTION_IDENTIFIED_FUNCTION to bind the correct
 * function interface identifier to a function definition.
 *
 * A technical note: we do not simply concatenate tokens with
 * the expansion of INTERFACE_IDENTIFIER_RAW because we wish
 * compilation to fail if INTERFACE_IDENTIFIER_RAW is undefined.
 */
#define VRPSX_IDENTIFIED_FUNCTION_NAME(PREFIX, INTERFACE)               \
    VRPSX_CAT2(INTERFACE, _IDENTIFIER_CAT)(VRPSX_CAT2(PREFIX, _),)

/* Expands to this function declaration:
 *
 *     STORAGE_CLASS INTERFACE_t
 *         VRPSX_IDENTIFIED_FUNCTION_NAME(PREFIX, INTERFACE)
 *
 * A typical STORAGE_CLASS is "static".
 *
 * The terminating semicolon is omitted to help smart editors;
 * you must append it.
 */
#define VRPSX_DECLARE_IDENTIFIED_FUNCTION(STORAGE_CLASS, PREFIX, INTERFACE) \
    STORAGE_CLASS VRPSX_CAT2(INTERFACE, _t)                             \
        VRPSX_IDENTIFIED_FUNCTION_NAME(PREFIX, INTERFACE)

/* Expands to this vrpsx_identified_function_t initializer:
 *
 *     { (void (*)(void)) & VRPSX_IDENTIFIED_FUNCTION_NAME(PREFIX, INTERFACE) \
 *     , INTERFACE_IDENTIFIER,
 *     , INTERFACE_DISPLAY_NAME }
 *
 * When used with VRPSX_DECLARE_IDENTIFIED_FUNCTION, ensure that you
 * bind the correct function interface identifier to the function.
 *
 * For example:
 *
 *     #include <vrpsx_system_main.h>
 *     #include <vrpsx_system_main_identifier.h>
 *     #include <vrpsx_identified_function.h>
 *     #include <vrpsx_identified_function_identifier.h>
 *     #include <vrpsx_utility.h>
 *
 *     VRPSX_DECLARE_IDENTIFIED_FUNCTION(static, main, vrpsx_system_main);
 *
 *     static int VRPSX_IDENTIFIED_FUNCTION_NAME(main, vrpsx_system_main)(
 *         vrpsx_system_t *system,
 *         int argc,
 *         char const *const *argv)
 *     {
 *         return VRPSX_CONSTANT(system, exit_success);
 *     }
 *
 *     vrpsx_identified_function_t const VRPSX_IDENTIFIED_FUNCTION_LIST[] = {
 *         VRPSX_MENTION_IDENTIFIED_FUNCTION(main, vrpsx_system_main)
 *     };
 *
 * expands to something like:
 *
 *     static vrpsx_system_main_t main_abcdef01;
 *
 *     static int main_abcdef01(
 *         vrpsx_system_t *system,
 *         int argc,
 *         char const *const *argv)
 *     {
 *         return ((system)->vtbl->exit_success);
 *     }
 *
 *     vrpsx_identified_function_t const _87654321[] = {
 *         { (void (*)(void)) & main_abcdef01
 *         , 0xabcdef01U,
 *         "vrpsx_system_main" }
 *     };
 *
 * but with different actual interface identifiers, probably.
 */
#define VRPSX_MENTION_IDENTIFIED_FUNCTION(PREFIX, INTERFACE)            \
    { (void (*)(void)) & VRPSX_IDENTIFIED_FUNCTION_NAME(PREFIX, INTERFACE) \
    , VRPSX_CAT2(INTERFACE, _IDENTIFIER)                                \
    , VRPSX_CAT2(INTERFACE, _DISPLAY_NAME) }

/* Expands to this vrpsx_identified_function_t initializer:
 *
 *     { NULL, 0, NULL }
 *
 * Use this initializer to terminate a list of identified functions,
 * as described in <vrpsx_identified_function.h>.
 */
#define VRPSX_IDENTIFIED_FUNCTION_LIST_TERMINATOR  { NULL, 0, NULL }


/* Initializes the "array" and "dimension" members of an vrpsx_argv_t.
 *
 * Naturally you should also initialize the "vtbl" member,
 * but currently that does not require the same kind error-prone
 * const-casting, and so it does not require a macro.  Besides,
 * sometimes that assignment occurs separately, at compile time.
 *
 * WARNING: Such initialization should occur only by the implementation
 * during interface construction, BEFORE there are any interface clients.
 * Otherwise the requirements of the interface would be violated.
 *
 * See the definition of vrpsx_argv_t for the full rules; this
 * macro is merely a convenience, NOT part of the interface.
 *
 * Evaluates INTERFACE, ARRAY, and DIMENSION exactly once.
 */
#define VRPSX_ARGV_INITIALIZE_ARRAY(INTERFACE, DIMENSION, ARRAY)  {     \
    vrpsx_argv_t *interface_5e698e28 = (INTERFACE);                     \
    char const *const *const *array_5e698e28 = &interface_5e698e28->array; \
    int const *dimension_5e698e28 = &interface_5e698e28->dimension;     \
    *(char ***)array_5e698e28 = (ARRAY);                                \
    *(int *)dimension_5e698e28 = (DIMENSION);                           \
}


/* Like memcpy(TARGET, SOURCE, LENGTH), but also adds LENGTH to TARGET.
 * TARGET must be an lvalue of type (char *).
 *
 * Evaluates TARGET, SOURCE, and LENGTH exactly once.
 */
#define MEMCPY_END(TARGET, SOURCE, LENGTH)  {                   \
    char **target_7568e5ae = &(TARGET);                         \
    void const *source_7568e5ae = (SOURCE);                     \
    size_t length_7568e5ae = (LENGTH);                          \
                                                                \
    memcpy(*target_7568e5ae, source_7568e5ae, length_7568e5ae); \
    *target_7568e5ae += length_7568e5ae;                        \
}


/* Writes one character.  Equivalent To:
 *
 *      {
 *          vrpsx_ostreambuf_t *osb = (OSB);
 *          char ch = (CH);
 *
 *          if (VRPSX_CALL1(osb, put1, ch)) {
 *              ERR;
 *          }
 *      }
 *
 * except that this macro may make direct use of the buffer for speed,
 * and that "osb" and "ch" are named in such a way as to minimize
 * the chance of collision with other variable names in your program.
 *
 * This optional macro is intended for convenience, not encapsulation;
 * it does not define any aspect of the vrpsx_ostreambuf_t interface.
 * Feel free to look at the definition of this macro.
 *
 * ERR can obtain the error number from the "error" member of OSB.
 *
 * Evaluates OSB and CH exactly once.  On error, executes ERR.
 */
#define VRPSX_OSTREAMBUF_PUT1(OSB, CH, ERR)  {                          \
    vrpsx_ostreambuf_t *osb_50d6c3bf00cbb826 = (OSB);                   \
    char ch_50d6c3bf00cbb826 = (CH);                                    \
                                                                        \
    if (osb_50d6c3bf00cbb826->current != osb_50d6c3bf00cbb826->end) {   \
        *osb_50d6c3bf00cbb826->current++ = ch_50d6c3bf00cbb826;         \
    } else if (VRPSX_CALL1(osb_50d6c3bf00cbb826,                        \
            put1, ch_50d6c3bf00cbb826)) {                               \
        ERR;                                                            \
    }                                                                   \
}

/* Writes several characters.  Equivalent To:
 *
 *      {
 *          vrpsx_ostreambuf_t *osb = (OSB);
 *          char const *begin = (BEGIN);
 *          size_t size = (SIZE);
 *
 *          if (VRPSX_CALL2(osb, putn, begin, size)) {
 *              ERR;
 *          }
 *      }
 *
 * except that this macro may make direct use of the buffer for speed,
 * and that "osb", "begin", and "size" are named in such a way as to minimize
 * the chance of collision with other variable names in your program.
 *
 * This optional macro is intended for convenience, not encapsulation;
 * it does not define any aspect of the vrpsx_ostreambuf_t interface.
 * Feel free to look at the definition of this macro.
 *
 * ERR can obtain the error number from the "error" member of OSB.
 *
 * Evaluates OSB, BEGIN, and SIZE exactly once.  On error, executes ERR.
 */
#define VRPSX_OSTREAMBUF_PUTN(OSB, BEGIN, SIZE, ERR)  {                 \
    vrpsx_ostreambuf_t *osb_14df186cd0a15d83 = (OSB);                   \
    char const *begin_14df186cd0a15d83 = (BEGIN);                       \
    size_t size_14df186cd0a15d83 = (SIZE);                              \
                                                                        \
    if (size_14df186cd0a15d83 < 8                                       \
          && size_14df186cd0a15d83                                      \
            <= osb_14df186cd0a15d83->end - osb_14df186cd0a15d83->current) { \
        memcpy(osb_14df186cd0a15d83->current,                           \
            begin_14df186cd0a15d83, size_14df186cd0a15d83);             \
        osb_14df186cd0a15d83->current += size_14df186cd0a15d83;         \
    } else if (VRPSX_CALL2(osb_14df186cd0a15d83,                        \
            putn, begin_14df186cd0a15d83, size_14df186cd0a15d83)){      \
        ERR;                                                            \
    }                                                                   \
}


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* VRPSX_UTILITY_H */
